"""Base type for objects supporting yakonfig.

.. This software is released under an MIT/X11 open source license.
   Copyright 2014-2015 Diffeo, Inc.

Purpose
=======

Provides a common place to declare command-line arguments and default
configuration to yakonfig.  Configurable object classes can be passed
into :func:`yakonfig.parse_args`.  This will cause yakonfig to
instantiate the relevant command-line arguments, parse any inbound
YAML configuration, fill in defaults, and produce a complete
configuration.

Only :attr:`Configurable.config_name` is strictly required, the
remainder of the functions can be absent or have defaults similar to
what is here.

Implementations of these methods may also find :func:`check_subconfig`
useful.

Implementation Details
======================

:func:`yakonfig.parse_args` doesn't actually require
:class:`Configurable` subclasses as its parameter; any object (e.g.,
module objects) that include the required names can be used.

Module Contents
===============

"""

from __future__ import absolute_import
import abc
import collections

from .exceptions import ProgrammerError

class Configurable(object):
    """Description of yakonfig configuration.

    This class provides a common place to define information that
    drives the configuration process.  Yakonfig doesn't actually
    require instances of this type, merely objects that provide the
    object and method names described here, and of these only
    :attr:`config_name` is truly required.

    For instance, if you create a subclass of this and then use
    yakonfig to configure it, yakonfig will fill in the default
    values described here

    >>> class MyConfigurable(object):
    ...     config_name = "me"
    ...     default_config = { 'key': 'value' }
    >>> yakonfig.set_default_config([MyConfigurable])
    >>> yaml.dump(yakonfig.get_global_config())
    me:
        key: value

    If you subclass this, you will need to pass instances of this
    object to the top-level yakonfig methods, not the class itself.  A
    corollary to this is that it is possible for the command-line
    arguments to vary based on parameters to this object's
    constructor.

    If you want command-line arguments to be able to affect the
    configuration this object describes, add them in
    :meth:`add_arguments()`, and set :attr:`runtime_keys` to a mapping
    from argparse name to config key name.

    """

    __metaclass__ = abc.ABCMeta

    @abc.abstractproperty
    def config_name(self):
        """Name of this configuration block."""
        return None

    @property
    def default_config(self):
        """Default values for configuration.

        This is a dictionary mapping configuration key name to configuration
        value.  It should not include any configured sub-objects.

        """
        return {}

    @property
    def services(self):
        """
        A list of services that a parent configuration can inject.
        """
        return []

    @property
    def sub_modules(self):
        """Modules this module controls.

        This is any iterable containing a sequence of
        :class:`Configurable` objects (or objects that act like them).
        Those configurations will be stored under this object's
        configuration, with the names specified by their
        :attr:`config_name` properties.

        """
        return []

    def add_arguments(self, parser):
        """Add additional command-line arguments to `parser`.

        :param parser: command-line argument parser
        :type parser: :class:`argparse.ArgumentParser`

        """
        pass

    @property
    def runtime_keys(self):
        """Mapping of argparse keys to configuration keys.

        This is used to capture the command-line arguments in
        :meth:`add_arguments`.  It is a dictionary mapping argparse
        argument name to configuration key name.

        """
        return {}

    def replace_config(self, config, name=''):
        '''Look at `config` and return a new :class:`Configurable`.

        This can, for instance, load external modules pointed to by
        a configuration file, and return a new :class:`ProxyConfigurable`
        for this with a new :attr:`sub_modules` list.  The yakonfig
        framework will not recursively call this function on submodules.

        :param dict config: configuration of this object and its children
        :param str name: name of the configuration block
        :return: replacement for `self`
        :rtype: :class:`Configurable`
        :raises yakonfig.ConfigurationError: if the new
          configuration cannot be generated
        '''
        return self

    def discover_config(self, config, name=''):
        '''Fill in missing values in `config` from the environment.

        This can literally look in :attr:`os.environ` to pick up values
        such as ``$TMPDIR``, or values published from a container system
        such as `Docker <http://www.docker.com/>`_.  It could also
        look at an external system, such as `Apache ZooKeeper
        <http://zookeeper.apache.org/>`_ or `CoreOS etcd
        <https://coreos.com/docs/distributed-configuration/etcd-api/>`_
        to find information such as what systems exist in the cluster.

        This method should limit itself to filling in missing values.
        It should modify `config` in place.  The framework will call
        this method once, walking the configurable tree for you.
        `config` will not have default values applied to it at this
        point.  If :meth:`replace_config` is also implemented, this
        method will be called only on the new object returned by
        :meth:`replace_config`.  Implementations of this method should
        be careful to trap exceptions; failure to reach an external
        configuration management system should not break initial
        configuration.

        :param dict config: configuration of this object and its
          children (read/write, does not include defaults)
        :param str name: name of the configuration block

        '''
        pass

    def check_config(self, config, name=''):
        """Validate the configuration of this object.

        If something is missing, incorrect, or inconsistent, raise a
        :exc:`yakonfig.ConfigurationError`.

        :param dict config: configuration of this object and its children
        :param str name: name of the configuration block, ending in
          :attr:`config_name`
        :raises yakonfig.ConfigurationError: if the
          configuration is invalid in some way

        """
        pass

    def normalize_config(self, config):
        """Rewrite the configuration of this object.

        This has the opportunity to, for example, replace paths with
        absolute paths, and to push configuration into sub-module
        configuration.  This may alter `config` in place.

        :param dict config: configuration of this object and its children

        """
        pass

class ProxyConfigurable(Configurable):
    '''A yakonfig configurable object that passes calls on to something else.

    The object this proxies can be any any configurable thing, not
    necessarily a :class:`Configurable` instance.  Any methods that
    are not implemented in the underlying object return default
    values.

    This class is intended to be used as a base class for
    :meth:`Configurable.replace_config` implementations.

    .. automethod:: __init__
    '''

    def __init__(self, config=None, *args, **kwargs):
        '''Create a new proxy configurable.

        :param config: object to pass options on to
        :type config: :class:`Configurable`

        '''
        super(ProxyConfigurable, self).__init__(*args, **kwargs)
        self.config = config

    def _property(self, name):
        if hasattr(self.config, name):
            return getattr(self.config, name)
        return getattr(super(ProxyConfigurable, self), name)

    @property
    def config_name(self):
        return self._property('config_name')

    @property
    def default_config(self):
        return self._property('default_config')

    @property
    def sub_modules(self):
        return self._property('sub_modules')

    def add_arguments(self, parser):
        if hasattr(self.config, 'add_arguments'):
            return getattr(self.config, 'add_arguments')(parser)
        return super(ProxyConfigurable, self).add_arguments(parser)

    @property
    def runtime_keys(self):
        return self._property('runtime_keys')

    def replace_config(self, config, name=''):
        if hasattr(self.config, 'replace_config'):
            return getattr(self.config, 'replace_config')(config, name)
        return super(ProxyConfigurable, self).replace_config(config, name)

    def discover_config(self, config, name=''):
        if hasattr(self.config, 'discover_config'):
            return getattr(self.config, 'discover_config')(config, name)
        return super(ProxyConfigurable, self).discover_config(config, name)

    def check_config(self, config, name=''):
        if hasattr(self.config, 'check_config'):
            return getattr(self.config, 'check_config')(config, name)
        return super(ProxyConfigurable, self).check_config(config, name)

    def normalize_config(self, config):
        if hasattr(self.config, 'normalize_config'):
            return getattr(self.config, 'normalize_config')(config)
        return super(ProxyConfigurable, self).normalize_config(config)

class NewSubModules(ProxyConfigurable):
    '''A proxy that only replaces the :attr:`sub_modules` list.

    This is expected to be the common use case for
    :meth:`replace_config()`.

    .. automethod: __init__
    '''
    def __init__(self, config=None, sub_modules=None, *args, **kwargs):
        '''Create a new proxy with new :attr:`sub_modules`.

        :param config: original `Configurable` object
        :type config: :class:`Configurable`
        :param sub_modules: new `sub_modules` list
        :type sub_modules: list of :class:`Configurable`

        '''
        super(NewSubModules, self).__init__(config=config, *args, **kwargs)
        self._sub_modules = sub_modules or []

    @property
    def sub_modules(self):
        return self._sub_modules

def check_subconfig(config, name, sub):
    """Validate the configuration of an object within this.

    This calls :meth:`Configurable.check_config` or equivalent on `sub`.
    A dictionary configuration for `sub` is required in `config`.

    >>> def check_config(config, name):
    ...     for sub in sub_modules:
    ...         check_subconfig(config, name, sub)

    :param dict config: parent configuration
    :param str name: name of the parent configuration block
    :param sub: Configurable-like subobject to check
    :raise yakonfig.ConfigurationError: if there is no
      configuration for `sub`, or it is not a dictionary

    """
    subname = sub.config_name
    subconfig = config.setdefault(subname, {})
    if not isinstance(subconfig, collections.Mapping):
        raise ProgrammerError('configuration for {} in {} must be a mapping'
                              .format(subname, name))
    checker = getattr(sub, 'check_config', None)
    if checker is not None:
        checker(subconfig, '{}.{}'.format(name, subname))



#!python

from __future__ import absolute_import, division, print_function
import argparse
import logging
import os
import tempfile

import pytest
from six import StringIO

from yakonfig import set_global_config, get_global_config, clear_global_config
## for use in fixture below
import yakonfig.yakonfig as yakonfig_internals

logger = logging.getLogger(__name__)

@pytest.fixture
def reset_globals(request):
    '''
    for fixture that makes each test run as if it were the first call
    to yakonfig
    '''
    clear_global_config()


def test_yakonfig_simple(reset_globals):
    YAML_TEXT_ONE = StringIO('''
pipeline_property1: run_fast
pipeline_property2: no_errors
''')
    config = set_global_config(YAML_TEXT_ONE)

    assert get_global_config() is config

    assert config['pipeline_property1'] == 'run_fast'
    assert config['pipeline_property2'] == 'no_errors'


@pytest.fixture
def monkeypatch_open(request):
    ## cannot pytest.monkeypatch a builtin like `open`, so instead,
    ## override method on Loader class
    def other_open(*args, **kwargs):
        fh = StringIO('''
k1: v1
k2: 
 - v21
''')
        fh.__exit__ = lambda x,y,z: None
        fh.__enter__ = lambda : fh
        return fh
    real_open = yakonfig_internals.Loader.open
    yakonfig_internals.Loader.open = other_open
    def fin():
        yakonfig_internals.Loader.open = real_open
    request.addfinalizer(fin)

def test_include_yaml_abstract(reset_globals, monkeypatch_open):
    YAML_TEXT_TWO = StringIO('''
app_one:
  one: car

app_two:
  bad: [cat, horse]
  good: !include_yaml /some-path-that-will-not-be-used
''')
    config = set_global_config(YAML_TEXT_TWO)
    
    assert get_global_config() is config
    sub_config = get_global_config('app_two')

    assert sub_config is config['app_two']
    assert sub_config['good'] == dict(k1='v1', k2=['v21'])

def test_include_abstract(reset_globals, monkeypatch_open):
    YAML_TEXT_TWO = StringIO('''
app_one:
  one: car

app_two:
  bad: [cat, horse]
  good: !include /some-path-that-will-not-be-used
''')
    config = set_global_config(YAML_TEXT_TWO)
    
    assert get_global_config() is config
    sub_config = get_global_config('app_two')

    assert sub_config is config['app_two']
    assert sub_config['good'] == dict(k1='v1', k2=['v21'])

def test_include_real_paths(reset_globals):
    t1 = tempfile.NamedTemporaryFile()
    t2 = tempfile.NamedTemporaryFile()
    t3 = tempfile.NamedTemporaryFile()
    y1 = u'''
t1:
  k3: !include_yaml %s
  k4: !include_yaml %s
''' % (t2.name, os.path.basename(t3.name))
    print(y1)
    y2 = u'dog'
    y3 = u'two'
    t1.write(y1.encode('utf-8'))
    t2.write(y2.encode('utf-8'))
    t3.write(y3.encode('utf-8'))
    t1.flush()
    t2.flush()
    t3.flush()

    config = set_global_config(t1.name)
    assert get_global_config() is config
    print(config)
    sub_config = get_global_config('t1')
    assert sub_config is config['t1']
    assert sub_config['k3'] == y2
    assert sub_config['k4'] == y3



# -*- coding: utf-8 -*-
# Author: Douglas Creager <dcreager@dcreager.net>
# This file is placed into the public domain.
 
# Calculates the current version number.  If possible, this is the
# output of “git describe”, modified to conform to the versioning
# scheme that setuptools uses.  If “git describe” returns an error
# (most likely because we're in an unpacked copy of a release tarball,
# rather than in a git working copy), then we fall back on reading the
# contents of the RELEASE-VERSION file.
#
# To use this script, simply import it your setup.py file, and use the
# results of get_git_version() as your package version:
#
# from version import get_git_version
#
# setup(
#     version=get_git_version()[0],
#     .
#     .
#     .
# )
#
# This will automatically update the RELEASE-VERSION file, if
# necessary.  Note that the RELEASE-VERSION file should *not* be
# checked into git; please add it to your top-level .gitignore file.
#
# You'll probably want to distribute the RELEASE-VERSION file in your
# sdist tarballs; to do this, just create a MANIFEST.in file that
# contains the following line:
#
#   include RELEASE-VERSION
from __future__ import absolute_import, division, print_function
 
__all__ = ("get_git_version")

import os
import sys
import traceback
from subprocess import Popen, PIPE
 
 
def call_git_describe(abbrev=4):
    dot_git = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        '.git')
    if not os.path.exists(dot_git):
        return None, None

    line = None
    p = None
    try:
        p = Popen(['git', 'describe', '--abbrev=%d' % abbrev],
                  stdout=PIPE, stderr=PIPE, 
                  cwd=os.path.dirname(os.path.abspath(__file__)),
                  universal_newlines=True)
        p.stderr.close()
        describe_line = p.stdout.readlines()[0].strip()

        p = Popen(['git', 'rev-parse', 'HEAD'],
                  stdout=PIPE, stderr=PIPE)
        p.stderr.close()
        source_hash = p.stdout.readlines()[0].strip()
        source_hash = source_hash[:abbrev]

        parts = describe_line.split('-')
        if len(parts) == 1:
            version = parts[0]

        else:
            ver, rel, source_hash = parts
            version_parts = ver.split('.')
            lasti = len(version_parts) - 1
            # increment whatever the last part of this a.b.c.d.yadda
            version_parts[lasti] = str(int(version_parts[lasti]) + 1)
            version = '{}.dev{}'.format('.'.join(version_parts), rel)

        return version, source_hash
 
    except Exception as exc:
        sys.stderr.write('line: %r\n' % line)
        sys.stderr.write(traceback.format_exc(exc))
        try:
            sys.stderr.write('p.stderr.read()=%s\n' % p.stderr.read())
        except Exception as exc:
            sys.stderr.write(traceback.format_exc(exc))
        try:
            sys.stderr.write('os.getcwd()=%s\n' % os.getcwd())
        except Exception as exc:
            sys.stderr.write(traceback.format_exc(exc))
        return None, None
 
 
def read_release_version():
    try:
        f = open("RELEASE-VERSION", "r")
 
        try:
            version = f.readlines()[0]
            return version.strip().split(',')
 
        finally:
            f.close()
 
    except:
        return None, None
 
 
def write_release_version(version, source_hash):
    f = open("RELEASE-VERSION", "w")
    f.write("%s,%s\n" % (version, source_hash))
    f.close()
 
 
def get_git_version(abbrev=4):
    # Read in the version that's currently in RELEASE-VERSION.
 
    release_version, release_source_hash = read_release_version()
 
    # First try to get the current version using “git describe”.
 
    version, source_hash = call_git_describe(abbrev)
 
    # If that doesn't work, fall back on the value that's in
    # RELEASE-VERSION.
 
    if version is None:
        version = release_version
        source_hash = release_source_hash
 
    # If we still don't have anything, that's an error.
 
    if version is None:
        # raise ValueError("Cannot find the version number!")
        version = '0.0.0'
        source_hash = ''
 
    # If the current version is different from what's in the
    # RELEASE-VERSION file, update the file to be current.
 
    if version != release_version or source_hash != release_source_hash:
        write_release_version(version, source_hash)
 
    # Finally, return the current version.
 
    return version, source_hash
 
 
if __name__ == "__main__":
    print(get_git_version())



'''Discover object configuration from parameter lists.

.. This software is released under an MIT/X11 open source license.
   Copyright 2014-2015 Diffeo, Inc.

.. autoclass:: AutoFactory

'''
from __future__ import absolute_import, division, print_function
import abc
import inspect

from six import iteritems, string_types

from yakonfig.configurable import Configurable
from yakonfig.exceptions import ConfigurationError, ProgrammerError


class AutoFactory(Configurable):
    '''A configurable that discovers the its childrens' configuration.

    Clients that subclass :class:`AutoFactory` must implement the
    :attr:`auto_config` property, which should be an iterable of
    things to automatically a configuration from. Notably, subclasses
    should *not* implement :attr:`~yakonfig.Configurable.sub_modules`,
    as this class provides its own implementation of it using
    :attr:`auto_config`.

    Items in :attr:`auto_config` may be functions, methods, or
    classes.  If a class, its `__new__` or `__init__` function will be
    inspected.  Any parameters beyond :keyword:`self` are extracted
    from this function.  Keyword parameters with default values become
    the default configuration of the object.  Parameters without
    default values must be provided by the :class:`AutoFactory`
    instance as properties of the derived class.  The configurable
    callables may not take variable argument lists (``*args,
    **kwargs``).  If they are classes, they may have a
    :attr:`~yakonfig.Configurable.config_name` but no other
    :class:`~yakonfig.Configurable` metadata.

    As a special case, if a configurable callable contains a parameter
    named `config`, that parameter will receive the child's config
    dictionary (as distinct from the parent's saved config dictionary
    :attr:`config`), and the configuration check will allow unexpected
    configuration values.  This supports older modules that explicitly
    passed configuration dictionaries to their children.  If the
    configurable callable is a class with a
    :attr:`~yakonfig.Configurable.config_name` then it is used
    directly during the configuration cycle.

    This class hooks into the :mod:`yakonfig` configuration sequence
    to capture its part of the configuration at startup time.  Child
    callables' configuration may not contain settings that are not any
    of the defaulted parameters.  If this class is used outside the
    standard sequence, then you must set :attr:`config` before calling
    :meth:`create`.

    Currently, this class does **not** support a hierarchical
    configuration: items in :attr:`auto_config` may not be
    :class:`~yakonfig.Configurable`, and :class:`AutoFactory` objects
    may not be nested.  This is likely to change in the future.

    '''
    __metaclass__ = abc.ABCMeta

    def __init__(self, config=None):
        '''Create a new object factory.

        If `config` is :const:`None` (the default), then the
        :attr:`config` property must be set before calling
        :meth:`create`.  Passing this factory object to
        :func:`yakonfig.parse_args` or a similar high-level
        :mod:`yakonfig` setup method will also accomplish this.

        :param dict config: local configuration dictionary (if available)

        '''
        super(AutoFactory, self).__init__()
        self._config = config

    @property
    def sub_modules(self):
        return [AutoConfigured.from_obj(obj, any_configurable=True)
                for obj in self.auto_config]

    @abc.abstractproperty
    def auto_config(self):
        '''Must return a list of objects to automatically configure.

        This list is interpreted shallowly. That is, all configuration
        from each object is discovered through its name and parameter
        list only. Everything else is ignored.

        '''
        pass

    def check_config(self, config, prefix=''):
        for child in self.sub_modules:
            if hasattr(child, 'check_config'):
                child.check_config(config)

    def normalize_config(self, config):
        '''Rewrite (and capture) the configuration of this object.'''
        self.config = config

    @property
    def config(self):
        '''Saved configuration for the factory and its sub-objects.'''
        if self._config is None:
            raise ProgrammerError(
                'Tried to access saved factory configuration before '
                'yakonfig configuration was run.')
        return self._config

    @config.setter
    def config(self, c):
        self._config = c
        self.new_config()

    def new_config(self):
        '''Hook called when the configuration changes.

        If factory implementations keep properties that are created
        from the configuration, this is a place to create or reset them.
        The base class implementation does nothing.

        '''
        pass

    def create(self, configurable, config=None, **kwargs):
        '''Create a sub-object of this factory.

        Instantiates the `configurable` object with the current saved
        :attr:`config`.  This essentially translates to
        ``configurable(**config)``, except services defined in the
        parent and requested by `configurable` (by setting the
        ``services`` attribute) are injected. If a service is not
        defined on this factory object, then a
        :exc:`yakonfig.ProgrammerError` is raised.

        If `config` is provided, it is a local configuration for
        `configurable`, and it overrides the saved local configuration
        (if any).  If not provided, then :attr:`config` must already be
        set, possibly by passing this object into the :mod:`yakonfig`
        top-level setup sequence.

        :param callable configurable: object to create
        :param dict config: local configuration for `configurable`
        :param kwargs: additional keyword parameters
        :return: ``configurable(**config)``

        '''
        # If we got passed a string, find the thing to make.
        if isinstance(configurable, string_types):
            candidates = [ac for ac in self.sub_modules
                          if ac.config_name == configurable]
            if len(candidates) == 0:
                raise KeyError(configurable)
            configurable = candidates[0]

        # Regenerate the configuration if need be.
        if not isinstance(configurable, AutoConfigured):
            configurable = AutoConfigured.from_obj(configurable)

        if config is None:
            config = self.config.get(configurable.config_name, {})

        # Iteratively build up the argument list.  If you explicitly
        # called this function with a config dictionary with extra
        # parameters, those will be lost.
        params = {}
        for other, default in iteritems(configurable.default_config):
            params[other] = kwargs.get(other, config.get(other, default))
        for other in getattr(configurable, 'services', []):
            # AutoConfigured.check_config() validates that this key
            # wasn't in the global config, so this must have come from
            # either our own config parameter, a keyword arg, or the
            # caller setting factory.config; trust those paths.
            if other == 'config':
                params[other] = dict(config, **kwargs)
            elif other in kwargs:
                params[other] = kwargs[other]
            elif other in config:
                params[other] = config[other]
            else:
                # We're not catching an `AttributeError` exception
                # here because it may case a net too wide which makes
                # debugging underlying errors more difficult.
                params[other] = getattr(self, other)
        return configurable(**params)


class AutoConfigured(Configurable):
    '''Configurable proxy object for callable children.

    This is an wrapper class that provides an implementation that
    satisfies :class:`yakonfig.Configurable` for objects that can have
    their configuration automatically discovered.

    This class is for internal library use only and is not part
    of the yakonfig factory API.

    '''
    def __init__(self, obj, config_name, services, default_config):
        #: Callable object this proxies.
        self.obj = obj
        self._config_name = config_name
        self._services = services
        self._default_config = default_config

    @classmethod
    def from_obj(cls, obj, any_configurable=False):
        '''Create a proxy object from a callable.

        If `any_configurable` is true, `obj` takes a parameter named
        ``config``, and `obj` smells like it implements
        :class:`yakonfig.Configurable` (it has a
        :attr:`~yakonfig.Configurable.config_name`), then return it
        directly.

        '''
        discovered = cls.inspect_obj(obj)
        if ((any_configurable and
             'config' in discovered['required'] and
             hasattr(obj, 'config_name'))):
            return obj
        return cls(obj, discovered['name'], discovered['required'],
                   discovered['defaults'])

    def __call__(self, *args, **kwargs):
        return self.obj(*args, **kwargs)

    @property
    def config_name(self):
        '''Name of this object as it appears in configuration.'''
        return self._config_name

    @property
    def services(self):
        '''List of externally provided parameter names.'''
        return self._services

    @property
    def default_config(self):
        '''Derived default configuration for this object.'''
        return self._default_config

    def check_config(self, config, name=''):
        '''Check that the configuration for this object is valid.

        This is a more restrictive check than for most :mod:`yakonfig`
        objects.  It will raise :exc:`yakonfig.ConfigurationError` if
        `config` contains any keys that are not in the underlying
        callable's parameter list (that is, extra unused configuration
        options).  This will also raise an exception if `config`
        contains keys that duplicate parameters that should be
        provided by the factory.

        .. note:: This last behavior is subject to change; future
                  versions of the library may allow configuration to
                  provide local configuration for a factory-provided
                  object.

        :param dict config: the parent configuration dictionary,
          probably contains :attr:`config_name` as a key
        :param str name: qualified name of this object in the configuration
        :raise: :exc:`yakonfig.ConfigurationError` if excess parameters exist

        '''
        # This is assuming that `config` is the config dictionary of
        # the *config parent*. That is, `config[self.config_name]`
        # exists.
        config = config.get(self.config_name, {})

        # Complain about additional parameters, unless this is an
        # older object that's expecting a config dictionary.
        extras = set(config.keys()).difference(self.default_config)
        if 'config' not in self.services and extras:
            raise ConfigurationError(
                'Unsupported config options for "%s": %s'
                % (self.config_name, ', '.join(extras)))

        # This only happens if you went out of your way to
        # circumvent the configuration and delete a parameter.
        missing = set(self.default_config).difference(config)
        if missing:
            raise ConfigurationError(
                'Missing config options for "%s": %s'
                % (self.config_name, ', '.join(missing)))

        # Did caller try to provide parameter(s) that we also expect
        # the factory to provide?
        duplicates = set(config.keys()).intersection(set(self.services))
        if duplicates:
            # N.B. I don't think the parameter can come from the
            # default config because Python will not let you have
            # `arg` and `arg=val` in the same parameter
            # list. (`discover_config`, below, guarantees that
            # positional and named parameters are disjoint.)
            raise ConfigurationError(
                'Disallowed config options for "%s": %s'
                % (self.config_name, ', '.join(duplicates)))

    @staticmethod
    def inspect_obj(obj):
        '''Learn what there is to be learned from our target.

        Given an object at `obj`, which must be a function, method or
        class, return a configuration *discovered* from the name of
        the object and its parameter list. This function is
        responsible for doing runtime reflection and providing
        understandable failure modes.

        The return value is a dictionary with three keys: ``name``,
        ``required`` and ``defaults``. ``name`` is the name of the
        function/method/class. ``required`` is a list of parameters
        *without* default values. ``defaults`` is a dictionary mapping
        parameter names to default values. The sets of parameter names in
        ``required`` and ``defaults`` are disjoint.

        When given a class, the parameters are taken from its ``__init__``
        method.

        Note that this function is purposefully conservative in the things
        that is will auto-configure. All of the following things will result
        in a :exc:`yakonfig.ProgrammerError` exception being raised:

        1. A parameter list that contains tuple unpacking. (This is invalid
           syntax in Python 3.)
        2. A parameter list that contains variable arguments (``*args``) or
           variable keyword words (``**kwargs``). This restriction forces
           an auto-configurable to explicitly state all configuration.

        Similarly, if given an object that isn't a function/method/class, a
        :exc:`yakonfig.ProgrammerError` will be raised.

        If reflection cannot be performed on ``obj``, then a ``TypeError``
        is raised.

        '''
        skip_params = 0
        if inspect.isfunction(obj):
            name = obj.__name__
            inspect_obj = obj
            skip_params = 0
        elif inspect.ismethod(obj):
            name = obj.im_func.__name__
            inspect_obj = obj
            skip_params = 1  # self
        elif inspect.isclass(obj):
            inspect_obj = None
            if hasattr(obj, '__dict__') and '__new__' in obj.__dict__:
                inspect_obj = obj.__new__
            elif hasattr(obj, '__init__'):
                inspect_obj = obj.__init__
            else:
                raise ProgrammerError(
                    'Class "%s" does not have a "__new__" or "__init__" '
                    'method, so it cannot be auto configured.' % str(obj))
            name = obj.__name__
            if hasattr(obj, 'config_name'):
                name = obj.config_name
            if not inspect.ismethod(inspect_obj) \
                    and not inspect.isfunction(inspect_obj):
                raise ProgrammerError(
                    '"%s.%s" is not a method/function (it is a "%s").'
                    % (str(obj), inspect_obj.__name__, type(inspect_obj)))
            skip_params = 1  # self
        else:
            raise ProgrammerError(
                'Expected a function, method or class to '
                'automatically configure, but got a "%s" '
                '(type: "%s").' % (repr(obj), type(obj)))

        argspec = inspect.getargspec(inspect_obj)
        if argspec.varargs is not None or argspec.keywords is not None:
            raise ProgrammerError(
                'The auto-configurable "%s" cannot contain '
                '"*args" or "**kwargs" in its list of '
                'parameters.' % repr(obj))
        if not all(isinstance(arg, string_types) for arg in argspec.args):
            raise ProgrammerError(
                'Expected an auto-configurable with no nested '
                'parameters, but "%s" seems to contain some '
                'tuple unpacking: "%s"'
                % (repr(obj), argspec.args))

        defaults = argspec.defaults or []
        # The index into `argspec.args` at which keyword arguments with default
        # values starts.
        i_defaults = len(argspec.args) - len(defaults)
        return {
            'name': name,
            'required': argspec.args[skip_params:i_defaults],
            'defaults': {k: defaults[i]
                         for i, k in enumerate(argspec.args[i_defaults:])},
        }



"""Yakonfig exception types.

.. This software is released under an MIT/X11 open source license.
   Copyright 2014 Diffeo, Inc.

"""

from __future__ import absolute_import

class ConfigurationError(Exception):
    """Some part of the user-provided configuration is incomplete or
    incorrect."""
    pass

class ProgrammerError(Exception):
    """Some part of the actual code does not meet requirements."""
    pass



"""A sample Configurable-like module.

-----

This software is released under an MIT/X11 open source license.

Copyright 2014 Diffeo, Inc.

"""

config_name = 'configurable'
default_config = { 'type': 'module' }



'''Top-level entry points to yakonfig.

.. This software is released under an MIT/X11 open source license.
   Copyright 2014-2015 Diffeo, Inc.

Most programs' `main()` functions will call yakonfig as::

    parser = argparse.ArgumentParser()
    yakonfig.parse_args(parser, [yakonfig, module, module...])

where the list of modules are top-level modules or other
:class:`yakonfig.Configurable` objects the program uses.

Test code and other things not driven by argparse can instead call::

    yakonfig.set_default_config([yakonfig, module, module, ...])

or::

    with yakofnig.defaulted_config([yakonfig, ...]):
      ...

.. autofunction:: parse_args
.. autofunction:: set_default_config
.. autofunction:: defaulted_config

'''

from __future__ import absolute_import
import collections
import contextlib
import copy
import sys

from six import iteritems, StringIO
import yaml as yaml_mod

from .exceptions import ConfigurationError, ProgrammerError
from .merge import overlay_config, diff_config
from .yakonfig import get_global_config, set_global_config, _temporary_config

# These implement the Configurable interface for yakonfig proper!
config_name = 'yakonfig'


def add_arguments(parser):
    '''Add command-line arguments for yakonfig proper.

    This is part of the :class:`~yakonfig.Configurable` interface, and
    is usually run by including :mod:`yakonfig` in the
    :func:`parse_args()` module list.

    :param argparse.ArgumentParser parser: command-line argument
      parser

    '''
    parser.add_argument('--config', '-c', metavar='FILE',
                        help='read configuration from FILE')
    parser.add_argument('--dump-config', metavar='WHAT', nargs='?',
                        help='dump out configuration then stop '
                        '(default, effective, full)')
runtime_keys = {'config': 'config'}


def parse_args(parser, modules, args=None):
    """Set up global configuration for command-line tools.

    `modules` is an iterable of
    :class:`yakonfig.Configurable` objects, or anything
    equivalently typed.  This function iterates through those objects
    and calls
    :meth:`~yakonfig.Configurable.add_arguments` on
    each to build up a complete list of command-line arguments, then
    calls :meth:`argparse.ArgumentParser.parse_args` to actually
    process the command line.  This produces a configuration that is a
    combination of all default values declared by all modules;
    configuration specified in ``--config`` arguments; and overriding
    configuration values specified in command-line arguments.

    This returns the :class:`argparse.Namespace` object, in case the
    application has defined its own command-line parameters and
    needs to process them.  The new global configuration can be
    obtained via :func:`yakonfig.get_global_config`.

    :param argparse.ArgumentParser parser: application-provided
      argument parser
    :param modules: modules or Configurable instances to use
    :type modules: iterable of :class:`~yakonfig.Configurable`
    :param args: command-line options, or `None` to use `sys.argv`
    :return: the new global configuration

    """
    collect_add_argparse(parser, modules)
    namespace = parser.parse_args(args)
    try:
        do_dump_config = getattr(namespace, 'dump_config', None)
        set_default_config(modules, params=vars(namespace),
                           validate=not do_dump_config)
        if do_dump_config:
            if namespace.dump_config == 'full':
                to_dump = get_global_config()
            elif namespace.dump_config == 'default':
                to_dump = assemble_default_config(modules)
            else:  # 'effective'
                to_dump = diff_config(assemble_default_config(modules),
                                      get_global_config())
            yaml_mod.dump(to_dump, sys.stdout)
            parser.exit()
    except ConfigurationError as e:
        parser.error(e)
    return namespace


def set_default_config(modules, params=None, yaml=None, filename=None,
                       config=None, validate=True):
    """Set up global configuration for tests and noninteractive tools.

    `modules` is an iterable of
    :class:`yakonfig.Configurable` objects, or anything
    equivalently typed.  This function iterates through those objects
    to produce a default configuration, reads `yaml` as though it were
    the configuration file, and fills in any values from `params` as
    though they were command-line arguments.  The resulting
    configuration is set as the global configuration.

    :param modules: modules or Configurable instances to use
    :type modules: iterable of :class:`~yakonfig.Configurable`
    :param dict params: dictionary of command-line argument key to values
    :param str yaml: global configuration file
    :param str filename: location of global configuration file
    :param dict config: global configuration object
    :param bool validate: check configuration after creating
    :return: the new global configuration
    :returntype: dict

    """
    if params is None:
        params = {}

    # Get the configuration from the file, or from params['config']
    file_config = {}
    if yaml is None and filename is None and config is None:
        if 'config' in params and params['config'] is not None:
            filename = params['config']
    if yaml is not None or filename is not None or config is not None:
        if yaml is not None:
            file_config = yaml_mod.load(StringIO(yaml))
        elif filename is not None:
            with open(filename, 'r') as f:
                file_config = yaml_mod.load(f)
        elif config is not None:
            file_config = config

    # First pass: set up to call replace_config()
    # Assemble the configuration from defaults + file + arguments
    base_config = copy.deepcopy(file_config)
    create_config_tree(base_config, modules)
    fill_in_arguments(base_config, modules, params)
    default_config = assemble_default_config(modules)
    base_config = overlay_config(default_config, base_config)

    # Replace the modules list (accommodate external modules)
    def replace_module(config, m):
        name = getattr(m, 'config_name')
        c = config.get(name, {})
        if hasattr(m, 'replace_config'):
            return getattr(m, 'replace_config')(c, name)
        return m
    modules = [replace_module(base_config, m) for m in modules]

    # Reassemble the configuration again, this time reaching out to
    # the environment
    base_config = file_config
    create_config_tree(base_config, modules)
    fill_in_arguments(base_config, modules, params)
    do_config_discovery(base_config, modules)
    default_config = assemble_default_config(modules)
    base_config = overlay_config(default_config, file_config)
    fill_in_arguments(base_config, modules, params)

    # Validate the configuration
    if validate and len(modules) > 0:
        mod = modules[-1]
        checker = getattr(mod, 'check_config', None)
        if checker is not None:
            with _temporary_config():
                set_global_config(base_config)
                checker(base_config[mod.config_name], mod.config_name)

    # All done, normalize and set the global configuration
    normalize_config(base_config, modules)
    set_global_config(base_config)
    return base_config


@contextlib.contextmanager
def defaulted_config(modules, params=None, yaml=None, filename=None,
                     config=None, validate=True):
    """Context manager version of :func:`set_default_config()`.

    Use this with a Python 'with' statement, like

    >>> config_yaml = '''
    ... toplevel:
    ...   param: value
    ... '''
    >>> with yakonfig.defaulted_config([toplevel], yaml=config_yaml) as config:
    ...    assert 'param' in config['toplevel']
    ...    assert yakonfig.get_global_config('toplevel', 'param') == 'value'

    On exit the global configuration is restored to its previous state
    (if any).

    :param modules: modules or Configurable instances to use
    :type modules: iterable of :class:`~yakonfig.Configurable`
    :param dict params: dictionary of command-line argument key to values
    :param str yaml: global configuration file
    :param str filename: location of global configuration file
    :param dict config: global configuration object
    :param bool validate: check configuration after creating
    :return: the new global configuration

    """
    with _temporary_config():
        set_default_config(modules, params=params, yaml=yaml,
                           filename=filename, config=config, validate=validate)
        yield get_global_config()


def check_toplevel_config(what, who):
    """Verify that some dependent configuration is present and correct.

    This will generally be called from a
    :meth:`~yakonfig.Configurable.check_config` implementation.
    `what` is a :class:`~yakonfig.Configurable`-like object.  If the
    corresponding configuration isn't present in the global
    configuration, raise a :exc:`yakonfig.ConfigurationError`
    explaining that `who` required it.  Otherwise call that module's
    :meth:`~yakonfig.Configurable.check_config` (if any).

    :param yakonfig.Configurable what: top-level module to require
    :param str who: name of the requiring module
    :raise yakonfig.ConfigurationError: if configuration for
      `what` is missing or incorrect

    """
    config_name = what.config_name
    config = get_global_config()
    if config_name not in config:
        raise ConfigurationError(
            '{} requires top-level configuration for {}'
            .format(who, config_name))
    checker = getattr(what, 'check_config', None)
    if checker:
        checker(config[config_name], config_name)


def _recurse_config(parent_config, modules, f, prefix=''):
    '''Walk through the module tree.

    This is a helper function for :func:`create_config_tree` and
    :func:`_walk_config`.  It calls `f` once for each module in the
    configuration tree with parameters `parent_config`, `config_name`,
    `prefix`, and `module`.  `parent_config[config_name]` may or may
    not exist (but could be populated, as :func:`create_config_tree`).
    If even the parent configuration doesn't exist, `parent_config`
    could be :const:`None`.

    :param dict parent_config: configuration dictionary holding
      configuration for `modules`, or maybe :const:`None`
    :param modules: modules or Configurable instances to use
    :type modules: iterable of :class:`~yakonfig.configurable.Configurable`
    :param f: callable to call on each module
    :param str prefix: prefix name of `parent_config`
    :return: `parent_config`

    '''
    for module in modules:
        config_name = getattr(module, 'config_name', None)
        if config_name is None:
            raise ProgrammerError('{!r} must provide a config_name'
                                  .format(module))
        new_name = prefix + config_name

        f(parent_config, config_name, new_name, module)

        _recurse_config((parent_config or {}).get(config_name, None),
                        getattr(module, 'sub_modules', []),
                        f,
                        new_name + '.')
    return parent_config


def create_config_tree(config, modules, prefix=''):
    '''Cause every possible configuration sub-dictionary to exist.

    This is intended to be called very early in the configuration
    sequence.  For each module, it checks that the corresponding
    configuration item exists in `config` and creates it as an empty
    dictionary if required, and then recurses into child
    configs/modules.

    :param dict config: configuration to populate
    :param modules: modules or Configurable instances to use
    :type modules: iterable of :class:`~yakonfig.configurable.Configurable`
    :param str prefix: prefix name of the config
    :return: `config`
    :raises yakonfig.ConfigurationError: if an expected name is present
      in the provided config, but that name is not a dictionary

    '''
    def work_in(parent_config, config_name, prefix, module):
        if config_name not in parent_config:
            # this is the usual, expected case
            parent_config[config_name] = {}
        elif not isinstance(parent_config[config_name], collections.Mapping):
            raise ConfigurationError(
                '{} must be an object configuration'.format(prefix))
        else:
            # config_name is a pre-existing dictionary in parent_config
            pass

    _recurse_config(config, modules, work_in)


def _walk_config(config, modules, f, prefix=''):
    """Recursively walk through a module list.

    For every module, calls ``f(config, module, name)`` where
    `config` is the configuration scoped to that module, `module`
    is the Configurable-like object, and `name` is the complete
    path (ending in the module name).

    :param dict config: configuration to walk and possibly update
    :param modules: modules or Configurable instances to use
    :type modules: iterable of :class:`~yakonfig.configurable.Configurable`
    :param f: callback function for each module
    :param str prefix: prefix name of the config
    :return: config

    """
    def work_in(parent_config, config_name, prefix, module):
        # create_config_tree() needs to have been called by now
        # and you should never hit either of these asserts
        if config_name not in parent_config:
            raise ProgrammerError('{} not present in configuration'
                                  .format(prefix))
        if not isinstance(parent_config[config_name], collections.Mapping):
            raise ConfigurationError(
                '{} must be an object configuration'.format(prefix))

        # do the work!
        f(parent_config[config_name], module, prefix)

    return _recurse_config(config, modules, work_in)


def collect_add_argparse(parser, modules):
    """Add all command-line options.

    `modules` is an iterable of
    :class:`yakonfig.configurable.Configurable` objects, or anything
    equivalently typed.  This calls
    :meth:`~yakonfig.configurable.Configurable.add_arguments` (if
    present) on all of them to set the global command-line arguments.

    :param argparse.ArgumentParser parser: argparse parser
    :param modules: modules or Configurable instances to use
    :type modules: iterable of :class:`~yakonfig.configurable.Configurable`

    """
    def work_in(parent_config, config_name, prefix, module):
        f = getattr(module, 'add_arguments', None)
        if f is not None:
            f(parser)
    _recurse_config(dict(), modules, work_in)
    return parser


def assemble_default_config(modules):

    """Build the default configuration from a set of modules.

    `modules` is an iterable of
    :class:`yakonfig.configurable.Configurable` objects, or anything
    equivalently typed.  This produces the default configuration from
    that list of modules.

    :param modules: modules or Configurable instances to use
    :type modules: iterable of :class:`~yakonfig.configurable.Configurable`
    :return: configuration dictionary

    """
    def work_in(parent_config, config_name, prefix, module):
        if config_name in parent_config:
            raise ProgrammerError('multiple modules providing {}'
                                  .format(prefix))
        parent_config[config_name] = dict(getattr(module, 'default_config',
                                                  {}))
    return _recurse_config(dict(), modules, work_in)


def fill_in_arguments(config, modules, args):
    """Fill in configuration fields from command-line arguments.

    `config` is a dictionary holding the initial configuration,
    probably the result of :func:`assemble_default_config`.  It reads
    through `modules`, and for each, fills in any configuration values
    that are provided in `args`.

    `config` is modified in place.  `args` may be either a dictionary
    or an object (as the result of :mod:`argparse`).

    :param dict config: configuration tree to update
    :param modules: modules or Configurable instances to use
    :type modules: iterable of :class:`~yakonfig.configurable.Configurable`
    :param args: command-line objects
    :paramtype args: dict or object
    :return: config

    """
    def work_in(config, module, name):
        rkeys = getattr(module, 'runtime_keys', {})
        for (attr, cname) in iteritems(rkeys):
            v = args.get(attr, None)
            if v is not None:
                config[cname] = v
    if not isinstance(args, collections.Mapping):
        args = vars(args)
    return _walk_config(config, modules, work_in)


def do_config_discovery(config, modules):
    '''Let modules detect additional configuration values.

    `config` is the initial dictionary with command-line and
    file-derived values, but nothing else, filled in.  This calls
    :meth:`yakonfig.configurable.Configurable.discover_config` on
    every configuration module.  It is expect that this method will
    modify the passed-in configuration dictionaries in place.

    :param dict config: configuration tree to update
    :param modules: modules or Configurable instances to use
    :type modules: iterable of :class:`~yakonfig.configurable.Configurable`
    :return: `config`

    '''
    def work_in(config, module, name):
        f = getattr(module, 'discover_config', None)
        if f:
            f(config, name)
    return _walk_config(config, modules, work_in)


def normalize_config(config, modules):
    """Normalize configuration values in the entire tree.

    `config` is a dictionary holding the almost-final configuration.
    Each module's
    :method:`yakonfig.configurable.Configurable.normalize_config`
    function is called to make changes such as pushing configuration
    into sub-module configuration blocks and making file paths
    absolute.

    :param dict config: configuration tree to update
    :param modules: modules or Configurable instances to use
    :type modules: iterable of :class:`~yakonfig.configurable.Configurable`

    """
    def work_in(config, module, name):
        f = getattr(module, 'normalize_config', None)
        if f:
            f(config)
    return _walk_config(config, modules, work_in)



#!/usr/bin/env python
import os
from setuptools import setup, find_packages
from version import get_git_version

VERSION, SOURCE_LABEL = get_git_version()
PROJECT = 'yakonfig'
AUTHOR = 'Diffeo, Inc.'
AUTHOR_EMAIL = 'support@diffeo.com'
URL = 'http://github.com/diffeo/yakonfig'
DESC = 'load a configuration dictionary for a large application'


def read_file(file_name):
    with open(os.path.join(os.path.dirname(__file__), file_name), 'r') as f:
        return f.read()

setup(
    name=PROJECT,
    version=VERSION,
    description=DESC,
    license=read_file('LICENSE.txt'),
    long_description=read_file('README.md'),
    # source_label=SOURCE_LABEL,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 2.7',
        'Development Status :: 3 - Alpha',
        'Topic :: Utilities',
        # MIT/X11 license http://opensource.org/licenses/MIT
        'License :: OSI Approved :: MIT License',
    ],
    install_requires=[
        'pyyaml',
        'six',
    ],
)



'''Global configuration management tools.

.. This software is released under an MIT/X11 open source license.
   Copyright 2013-2014 Diffeo, Inc.

.. only:: pyapi or all

    Loads a YAML file (with extensions) and makes it available
    globally to a Python application.  Application code can call
    :func:`yakonfig.get_global_config` to get values from the
    configuration store.

    In typical use, a program will declare some number of top-level
    configuration modules.  These modules mimic the structure of
    :class:`yakonfig.Configurable`, and are passed in directly to
    :func:`yakonfig.parse_args()`.  This makes the corresponding
    configuration available in the global configuration.  For
    example::

        # I am a_module.py
        import yakonfig

        # Configurable metadata
        config_name = 'a_module'
        default_config = { 'message': 'hello world' }
        def add_arguments(parser):
            parser.add_argument('--message')
        runtime_keys = { 'message': 'message' }

        def main():
            parser = argparse.ArgumentParser()
            args = yakonfig.parse_args(parser, [yakonfig, a_module])
            print yakonfig.get_global_config('a_module', 'message')
        if __name__ == '__main__':
            main()

    Running ``python -m a_module`` will print out "hello world";
    running ``python -m a_module --message goodbye`` will inject the
    "goodbye" message into the configuration, and then print it out.

    .. note:: Avoid using :meth:`argparse.ArgumentParser.set_defaults`
              in combination with :func:`yakonfig.parse_args`.  Any
              default values set this way will override values set in
              the user's configuration file.

    Most of the useful functions and objects in the submodules are
    re-exported from the top-level module.

.. only:: all

    Common command-line arguments
    =============================

    All programs based on :mod:`yakonfig` support the following
    command-line arguments.

.. only:: cliref or all

    .. program:: yakonfig

    .. option:: --config <file.yaml>, -c <file.yaml>

        Load `file.yaml` as the primary configuration file.  Options
        set in the configuration file override any default values;
        options set on the command line override values in this
        configuration file.

    .. option:: --dump-config {default|effective|full}

        Write the configuration as YAML to standard output, then stop
        immediately.  The argument indicates which configuration is
        dumped: if ``default`` then the default configuration,
        ignoring all user setup, is dumped; if ``full`` then the dump
        contains the complete configuration, including all default
        settings, settings from the configuration file, and
        command-line options; and if ``effective`` then the dump
        contains only those settings that differ from the defaults,
        producing the minimum configuration to recreate the current
        settings.

.. only:: pyapi or all

    YAML extensions
    ===============

    ``key: !include path``
      Loads a yaml file at `path` and inserts it as the value
      associated with `key`.

    Top-level entry points
    ======================

    .. autofunction:: parse_args
    .. autofunction:: set_default_config
    .. autofunction:: defaulted_config
    .. autofunction:: get_global_config
    .. autofunction:: set_global_config
    .. autofunction:: clear_global_config

    Configurable modules
    ====================

    Modules that can be configured by the user implement the
    :class:`yakonfig.Configurable` interface, or more often, have their
    classes and modules provide the same names.  Anything that includes a
    :attr:`~yakonfig.Configurable.config_name` property can be passed into
    the top-level entry points above.

    .. autoclass:: Configurable
       :members:

    .. autoclass:: ProxyConfigurable
       :members:

    .. autoclass:: NewSubModules
       :members:

    .. autofunction:: check_toplevel_config
    .. autofunction:: check_subconfig

    Operations on configuration
    ===========================

    Configuration is always passed around as plain Python dictionaries.
    If you can guarantee that a configuration has passed through yakonfig,
    then you can generally guarantee that its default values are present
    and that any defined checks have passed.

    .. autofunction:: overlay_config
    .. autofunction:: diff_config

    Exceptions
    ==========

    .. autoclass:: ConfigurationError
    .. autoclass:: ProgrammerError

    Interactive programs
    ====================

    .. currentmodule:: yakonfig.cmd
    .. autoclass:: ArgParseCmd
    .. currentmodule:: yakonfig

'''
from __future__ import absolute_import
from yakonfig.configurable import Configurable, ProxyConfigurable, \
    NewSubModules, check_subconfig
from yakonfig.exceptions import ConfigurationError, ProgrammerError
from yakonfig.merge import diff_config, overlay_config
from yakonfig.toplevel import parse_args, set_default_config, \
    defaulted_config, check_toplevel_config, \
    config_name, add_arguments, runtime_keys
from yakonfig.yakonfig import clear_global_config, set_global_config, \
    get_global_config
